package com.yonge.netty.dto;

import java.math.BigDecimal;

/**
 * 一个音符的频率范围，包含最大值和最小值
 */
public class NoteFrequencyRange {

	private double minFrequency;

	private double maxFrequency;

	public NoteFrequencyRange(double standardFrequecy, double frequency) {
		int midiNoteSize = 128;
		double[] midiNoteFrequencies = new double[midiNoteSize];

		for (int x = 0; x < midiNoteSize; ++x) {
			midiNoteFrequencies[x] = new BigDecimal(standardFrequecy).multiply(
					new BigDecimal(Math.pow(2, new BigDecimal(x - 69).divide(new BigDecimal(12), 6, BigDecimal.ROUND_HALF_UP).doubleValue()))).doubleValue();

			if (midiNoteFrequencies[x] >= frequency) {
				if (midiNoteFrequencies[x] - frequency > frequency - midiNoteFrequencies[x - 1]) {
					// frequency演奏的是上一个音符
					maxFrequency = midiNoteFrequencies[x - 1] + (midiNoteFrequencies[x] - midiNoteFrequencies[x - 1]) / 2;
					minFrequency = midiNoteFrequencies[x - 1] - (midiNoteFrequencies[x - 1] - midiNoteFrequencies[x - 2]) / 2;
				} else {
					// frequency演奏的是当前音符
					midiNoteFrequencies[x + 1] = new BigDecimal(standardFrequecy).multiply(
							new BigDecimal(Math.pow(2, new BigDecimal((x + 1) - 69).divide(new BigDecimal(12), 6, BigDecimal.ROUND_HALF_UP).doubleValue())))
							.doubleValue();
					maxFrequency = midiNoteFrequencies[x] + (midiNoteFrequencies[x + 1] - midiNoteFrequencies[x]) / 2;
					minFrequency = midiNoteFrequencies[x] - (midiNoteFrequencies[x] - midiNoteFrequencies[x - 1]) / 2;
				}
				break;
			}
		}
	}

	public NoteFrequencyRange(double frequency) {
		new NoteFrequencyRange(442, frequency);
	}

	public double getMinFrequency() {
		return minFrequency;
	}

	public void setMinFrequency(double minFrequency) {
		this.minFrequency = minFrequency;
	}

	public double getMaxFrequency() {
		return maxFrequency;
	}

	public void setMaxFrequency(double maxFrequency) {
		this.maxFrequency = maxFrequency;
	}

	@Override
	public boolean equals(Object obj) {
		if (obj instanceof NoteFrequencyRange) {
			NoteFrequencyRange nfr = (NoteFrequencyRange) obj;
			return this.minFrequency == nfr.minFrequency && this.maxFrequency == nfr.maxFrequency;
		}
		return false;
	}

}
